/*
Language: Visual Studio .NET 2002, mixed C/C++ with Win 32 API.

Description:

Displayed information
---------------------

Win32
-----
This program uses the Win32 API DeviceIOControl function with SCSI PASS THROUGH
DIRECT. Because of this it only works on Win NT4/2K/XP/2003 with full administrator
rights. Many of the functions, constants & data types r from Microsoft Windows.h.

Usage:
Go into command prompt and type in:
media_info <drive letter>

History
-------

v1.00 - First release
*/

#include <windows.h>
#include <stdio.h>
#include <string.h>


#include "my_ntddscsi.h"
#include "sam.h"
#include "spc.h"
#include "crc32.h"
#include "SectorTools.h"
#include "cdtext.h"

const int SectorsPerRead=6;
//Global variables..
T_SPDT_SBUF sptd_sb;  //SCSI Pass Through Direct variable.
UCHAR DataBuf[SectorsPerRead*2448];  //Buffer for holding data to/from drive.
LPTSTR szVolumeFormat = TEXT("\\\\.\\%c:");
LPTSTR szRootFormat = TEXT("%c:\\");


FILE  *OutputFileStream;


// Toc Info


typedef struct
{

    byte TrackNumber;
    long int  StartSectorLBA;
    long int  EndSectorLBA;

} TocStructure;

TocStructure TOC[100];

//long int TOC_Sectors[100];
//long int TOC_Track_Nos[100];


//Global variables
SCSI_PASS_THROUGH_DIRECT sptd;


// tempory variables for storing Q data Items


unsigned long maincrc;


HANDLE OpenVolume(TCHAR cDriveLetter)
{
    HANDLE hVolume;
    UINT uDriveType;
    TCHAR szVolumeName[8];
    TCHAR szRootName[5];
    DWORD dwAccessFlags;

    //printf("Opening a handle to the drive..");

    wsprintf(szRootName, szRootFormat, cDriveLetter);
    uDriveType = GetDriveType(szRootName);
    
    switch(uDriveType) {
    case DRIVE_CDROM:
        dwAccessFlags = GENERIC_READ | GENERIC_WRITE;
        break;
    default:
        //printf("done.\nDrive is not CDROM type.\n");
        return INVALID_HANDLE_VALUE;
    }
	
	//Format drive letter to e.g. \\.\D:
    wsprintf(szVolumeName, szVolumeFormat, cDriveLetter);
    
    //This will only work for CD/DVD device on W2K and higher.
    hVolume = CreateFile(szVolumeName,
                            dwAccessFlags,
                                FILE_SHARE_READ | FILE_SHARE_WRITE,
                                    NULL,
                                        OPEN_EXISTING,
                                            0,
                                            NULL );

    if (hVolume == INVALID_HANDLE_VALUE)
    {
        //Try again for NT4..
        dwAccessFlags = GENERIC_READ;
        hVolume = CreateFile(szVolumeName,
                                dwAccessFlags,
                                    FILE_SHARE_READ | FILE_SHARE_WRITE,
                                        NULL,
                                            OPEN_EXISTING,
                                                FILE_ATTRIBUTE_NORMAL,
                                                    NULL);
        if (hVolume == INVALID_HANDLE_VALUE)
        {
            //printf("failed.\n");
            printf("Could not create handle for CD/DVD device.\n");
        }
    }
    else
    {
        //printf("done.\n");
    }

    return hVolume;
}

BOOL CloseVolume(HANDLE hVolume)
{
    BOOL success;
    //printf("Closing handle to the drive..");
    success=CloseHandle(hVolume);
    //printf("done.\n");

    return success;

}

BOOL SendCommandToDrive(HANDLE hVolume,
                            UCHAR CMDLength, 
                                    UCHAR DataIn ,
                                        ULONG ReturnedDataSize, 
                                            ULONG TimeOutValue,
                                                PVOID PtrToRetBuffer,
                                                       unsigned char *CMD )

{
    DWORD dwBytesReturned;

    sptd_sb.sptd.Length=sizeof(SCSI_PASS_THROUGH_DIRECT);
    sptd_sb.sptd.PathId=0;   //SCSI card ID will be filled in automatically
    sptd_sb.sptd.TargetId=0; //SCSI target ID will also be filled in
    sptd_sb.sptd.Lun=0;      //SCSI lun ID will also be filled in
    sptd_sb.sptd.CdbLength=CMDLength; //CDB size is 12 for ReadCD MMC1 command
    sptd_sb.sptd.SenseInfoLength=MAX_SENSE_LEN; //The Sense Buffer 
    sptd_sb.sptd.DataIn=DataIn; //There will be data from drive
    sptd_sb.sptd.DataTransferLength=ReturnedDataSize; //Size of data - in this case 1 sector of data
    sptd_sb.sptd.TimeOutValue=TimeOutValue;  //SCSI timeout value (60 seconds - maybe it should be longer)
    sptd_sb.sptd.DataBuffer=PtrToRetBuffer;//(PVOID)&(DataBuf);//(PVOID)&(DatBuf);
    sptd_sb.sptd.SenseInfoOffset=sizeof(SCSI_PASS_THROUGH_DIRECT);
    ZeroMemory(sptd_sb.SenseBuf, MAX_SENSE_LEN);

    for (UCHAR Idx=0;
            Idx<16; 
                Idx++)
    {
        if (Idx<CMDLength)
        {
            sptd_sb.sptd.Cdb[Idx]=CMD[Idx];
        }
        else
        {
            sptd_sb.sptd.Cdb[Idx]=0;
        };
    };

    
    //Send the command to drive
    return DeviceIoControl(hVolume,
            IOCTL_SCSI_PASS_THROUGH_DIRECT,               
                (PVOID)&sptd_sb, 
                    (DWORD)sizeof(sptd_sb),
                           (PVOID)&sptd_sb,
                                (DWORD)sizeof(sptd_sb),                                                         
                                        &dwBytesReturned,                                    
                                                NULL);
}

const int RawDataLength=2352;
const int RawSubCodeLength=96;


BOOL ReadCDSectorsAndRAWPtoWSubs(HANDLE hVolume, 
                                    long int StartSector, 
                                        byte NumberOfSectors)
{
/* 
        READ CD command (0xBE) 
        + * byte 1: bit 0: 
        + *   - mmc1-mmc4: RELADR: must be 0 
        + *   - mmc5-    : obsolete 
        + * byte 1: bit 1: 
        + *   - mmc1-mmc3: Reserved 
        + *   - mmc4-    : Digital Audio Play: When set to 1 the drive will do things 
        + *                like mute and interpolate on errors 
        + * byte 1: bit 4-2: sector type 
        + *   - 000: All types (mandotary) 
        + *   - 001: CD-DA (mmc1-mmc3: optional, mmc4-: mandotary) 
        + *   - 002: Mode 1 (mandotary) 
        + *   - 003: Mode 2 formless (mmc1-mmc3: madotary, mmc4-: optional) 
        + *   - 004: Mode 2 form 1 (mandotary) 
        + *   - 005: Mode 2 form 2 (mandotary) 
        + * byte 2-5: Start LBA 
        + * byte 6-8: Length 
        + * byte 9: bit 0: reserved 
        + * byte 9: bit 2-1: C2 error info: 
        + *   - 00: No C2 info 
        + *   - 01: 1 bit for all 2352 bytes 
        + *   - 10: Bitwise or of C2 bytes + pad + 1 bit for all 2352 bytes. 
        + * byte 9: bit 3: EDC & ECC 
        + * byte 9: bit 4: User Data 
        + * byte 9: bit 6-5: Header Codes 
        + *   - 00: No header info 
        + *   - 01: 4 byte sector header 
        + *   - 10: 8 byte sector sub-header (mode 2) 
        + *   - 11: both header and sub-header 
        + * byte 9: bit 7: Sync 
        + * byte 10: bit 2-0: sub channel selection: 
        + *   - 000: No sub channel 
        + *   - 001: RAW P-W channel (only mentioned in mmc1 - mmc4, but others still 
        + *          talk about it.) 
        + *   - 010: Q channel data 
        + *   - 100: Corrected and deinterleaved R-W subchannel 
        + * 
        + * Note: Setting any of the bits 7 - 3 from byte 9 for should act the same 
        + *       and only return the user data (0x10) for CDDA. 
        + * 
 */ 
    unsigned char cmd[12] ={0xBE,
                        0,
                            (StartSector>>24) & 0xFF,
                                (StartSector>>16) & 0xFF,
                                    (StartSector>>8) & 0xFF,
                                        (StartSector) & 0xFF,
                                            0,
                                                0,
                                                    NumberOfSectors,
                                                        0xF8,
                                                            0x01,
                                                                0}; 

/*
    //CDB with values for ReadCD CDB12 command.  The values were taken from MMC1 draft paper.
    sptd.Cdb[0]=0xBE;  //Code for ReadCD CDB12 command
    sptd.Cdb[1]=0;        
    //Read upto 10 Seconds of each track to find 2 Successive Identical ISRCs 
    sptd.Cdb[5]=byte(StartSector);   //Least sig byte of LBA sector no. to read from CD
    StartSector=StartSector>>8;
    sptd.Cdb[4]=byte(StartSector);  //2nd byte of:
    StartSector=StartSector>>8;
    sptd.Cdb[3]=byte(StartSector);  //3rd byte of:
    StartSector=StartSector>>8;
    sptd.Cdb[2]=byte(StartSector);  //Most significant byte    
    sptd.Cdb[6]=0;  //Most significant byte of:
    sptd.Cdb[7]=0;  //Middle byte of:
    sptd.Cdb[8]=(NumberOfSectors & 255);  //Least sig byte of no. of sectors to read from CD
    sptd.Cdb[9]=0xF8;  //Raw read, 2352 bytes per sector
    sptd.Cdb[10]=1;  //Selects read RAW 96 bytes/sector sub-channel data
    sptd.Cdb[11]=0;
    sptd.Cdb[12]=0;
    sptd.Cdb[13]=0;
    sptd.Cdb[14]=0;
    sptd.Cdb[15]=0;
*/



    ULONG DataSize=(RawDataLength +RawSubCodeLength)*NumberOfSectors;
    return SendCommandToDrive( hVolume,
                            12, 
                                SCSI_IOCTL_DATA_IN , 
                                    DataSize, 
                                        100,
                                            &DataBuf,
                                                    cmd );
}



BOOL ReadISRCorUPC(HANDLE hVolume, 
              long int StartingSector, 
                    byte NumberOfSectorsPerRead, 
                        long int TotalSectorsToRead,bool FindISRC)
                      
{

    byte ErrorCount=0;
    long int CurrentSector=StartingSector;
    bool FoundValidItem=false;
    byte ItemToFind=2;
    
    if (FindISRC)
    {
        ItemToFind=3;
    }
   
    byte SectorsPerSample=NumberOfSectorsPerRead;

    while ((SectorsPerSample>0) && !(FoundValidItem) && ErrorCount<100)
    {

        if (ReadCDSectorsAndRAWPtoWSubs(hVolume,
                CurrentSector,
                    SectorsPerSample))
            {
                FoundValidItem=FindUPCorISRC(SectorsPerSample,ItemToFind,DataBuf);
                CurrentSector+=SectorsPerSample;
            }
        else
            if (ReadCDSectorsAndRAWPtoWSubs(hVolume,
                    CurrentSector,
                        1))
            {
                FoundValidItem=FindUPCorISRC(1,ItemToFind,DataBuf);
                CurrentSector++;
            }
            else
            {
                ErrorCount++;    
            };
        if ((StartingSector+TotalSectorsToRead)<(SectorsPerSample+CurrentSector))
        {
            SectorsPerSample=byte((StartingSector+TotalSectorsToRead)-CurrentSector);
        };

    };
    return FoundValidItem;
}


void WriteSectorsToFile(const byte SectorToOutput)
{   
    int i=0;
    for(i=0;(i<2352);i++)
    {    
         //   printf("%02X",DataBuf[i+(SectorToOutput*2448)]) ;
        //putc(DataBuf[i+(SectorToOutput*2448)],OutputFileStream);
        maincrc=crc32_addchar(DataBuf[i+(SectorToOutput*2448)],maincrc);
    };
    fwrite(&DataBuf[(SectorToOutput*2448)],1,2352,OutputFileStream); 
}


BOOL ReadSectorsDisk(HANDLE hVolume, 
              long int StartingSector, 
                    byte NumberOfSectorsPerRead, 
                        long int TotalSectorsToRead )
{
    int ErrorCount=0;  
    int PercentComplete=0;
    long int CurrentSector=StartingSector;
    byte SectorsPerSample=NumberOfSectorsPerRead;

    while ((SectorsPerSample>0) && ErrorCount<100)
    {
        if (ReadCDSectorsAndRAWPtoWSubs(hVolume,
                CurrentSector,
                    SectorsPerSample))
        {
            
            for (byte ind=0;(ind<SectorsPerSample);ind++)
            {
                //DecodeQData(ind);
                WriteSectorsToFile(ind);
                DetectSilence(ind,DataBuf);
            };
            CurrentSector+=SectorsPerSample;
        }
        else
            if (ReadCDSectorsAndRAWPtoWSubs(hVolume,
                    CurrentSector,
                        1))
            {
                //DecodeQData(0);
                WriteSectorsToFile(0);
                DetectSilence (0,DataBuf);
                CurrentSector++;
            }
            else
            {
                ErrorCount++;    
            };
        if ((StartingSector+TotalSectorsToRead)<(SectorsPerSample+CurrentSector))
        {
            SectorsPerSample=byte((StartingSector+TotalSectorsToRead)-CurrentSector);
        };

        if (int(((CurrentSector-StartingSector)*100)/TotalSectorsToRead)!=PercentComplete)
        {
            PercentComplete=int(((CurrentSector-StartingSector)*100)/TotalSectorsToRead);
            printf("%d\%\n",PercentComplete);
            fflush(stdout);
        };   
      
    };
    return (ErrorCount<10);
}


void OutputLongAs4ByteLittleEndienToFile(long int ToOutput)
{
    putc( ToOutput % 256 ,OutputFileStream);
    ToOutput=ToOutput>>8;
    putc( ToOutput % 256 ,OutputFileStream);
    ToOutput=ToOutput>>8;
    putc( ToOutput % 256 ,OutputFileStream);
    ToOutput=ToOutput>>8;
    putc( ToOutput % 256 ,OutputFileStream);
}


unsigned char *scsi_inquiry(HANDLE hVolume)
{

    unsigned char cmd[]={0x12,0,0,0,56,0};
    bool success=SendCommandToDrive(hVolume ,
                            6,
                                SCSI_IOCTL_DATA_IN,
                                    1000, //Size of return buffer 
                                        100, 
                                            &DataBuf,
                                               cmd);
        
    unsigned char *Target;
    Target=new unsigned char[50];
    memset(Target,0,50);
    if (success)
    {
        int count=0;
        while (DataBuf[8+count]!=0)
        {
            Target[count]=DataBuf[8+count];
            count++;
        };
    };

    return Target;
}



void ReadTrackToOutputStream(HANDLE hVolume,byte Track, char * FileName)
{

    errno_t err;
    printf("Opening - '" );
    printf("%s",FileName);
    printf("'\n");

    err=fopen_s(&OutputFileStream,(FileName) , "wb");

    //IsSilent =true;

    if (err==0)

    {
          
        long int SizeOfRIFF;
        long int SizeOfdata;
        long int SizeOfFormat=16;

        SizeOfdata = (TOC[Track].EndSectorLBA - TOC[Track].StartSectorLBA)*2352;
        SizeOfRIFF = 4+4+4+SizeOfFormat+4+4+SizeOfdata;


        //create RIFF HEADER
 

        putc( 'R' ,OutputFileStream);
        putc( 'I' ,OutputFileStream);
        putc( 'F' ,OutputFileStream);
        putc( 'F' ,OutputFileStream);
        OutputLongAs4ByteLittleEndienToFile(SizeOfRIFF); 

        putc( 'W' ,OutputFileStream);
        putc( 'A' ,OutputFileStream);
        putc( 'V' ,OutputFileStream);
        putc( 'E' ,OutputFileStream);
        
        putc( 'f' ,OutputFileStream);
        putc( 'm' ,OutputFileStream);
        putc( 't' ,OutputFileStream);
        putc( ' ' ,OutputFileStream);
        OutputLongAs4ByteLittleEndienToFile(SizeOfFormat);
        putc( 1 ,OutputFileStream);
        putc( 0 ,OutputFileStream);
        putc( 2 ,OutputFileStream); //channels
        putc( 0 ,OutputFileStream);
        OutputLongAs4ByteLittleEndienToFile(44100); //SampleRate
        OutputLongAs4ByteLittleEndienToFile(44100*4); //BytesPerSecond
        putc( 4 ,OutputFileStream); // bytes per block
        putc( 0 ,OutputFileStream);
        putc( 16 ,OutputFileStream); // bits per sample
        putc( 0 ,OutputFileStream);

        putc( 'd' ,OutputFileStream);
        putc( 'a' ,OutputFileStream);
        putc( 't' ,OutputFileStream);
        putc( 'a' ,OutputFileStream);
        OutputLongAs4ByteLittleEndienToFile(SizeOfdata);


        maincrc=crc32_initialise();
     
        ReadSectorsDisk(hVolume, 
                  TOC[Track].StartSectorLBA , 
                        SectorsPerRead, 
                            TOC[Track].EndSectorLBA - TOC[Track].StartSectorLBA );

        printf("CRC : %08x\n",crc32_read(maincrc));
        printf("Lead In Ends : %08d\n",GetLeadInEnd());
        printf("Lead Out Starts : %08d\n",GetLeadOutStart());
        fclose(OutputFileStream);

    }
    else
    {
        printf("Could not open file");
    };

};

void GenerateCRC32ForTrack(char * FileName, const long DriveOffset )
{

    long int SizeOfRIFF;
    long int SizeOfdata;
    long int Samples=0;
    long int SampleOffset;

    errno_t err;
    printf("Opening - '" );
    printf("%s",FileName);
    printf("'\n");

    err=fopen_s(&OutputFileStream,(FileName) , "rb");

    int GetChar;

    const int BufferSize=2048; // one frame
    unsigned char Buffer[BufferSize];

    if (err==0)

    {
        if ((getc(OutputFileStream)=='R') 
            && (getc(OutputFileStream)=='I') 
                && (getc(OutputFileStream)=='F')
                    && (getc(OutputFileStream)=='F'))

        {
            printf("RIFF Tag Found.......\n");
            SizeOfRIFF=getc(OutputFileStream);
            SizeOfRIFF+=getc(OutputFileStream)<<8;
            SizeOfRIFF+=getc(OutputFileStream)<<16;
            SizeOfRIFF+=getc(OutputFileStream)<<24;
            printf("RIFF Size . %12d .\n",SizeOfRIFF);

            for(int ind=0;ind<28;ind++)
            {
                GetChar=getc(OutputFileStream);
            }
            if ((getc(OutputFileStream)=='d') 
                && (getc(OutputFileStream)=='a') 
                    && (getc(OutputFileStream)=='t')
                        && (getc(OutputFileStream)=='a'))

            {        
                printf("data Tag Found.......\n");
                SizeOfdata=getc(OutputFileStream);
                SizeOfdata+=getc(OutputFileStream)<<8;
                SizeOfdata+=getc(OutputFileStream)<<16;
                SizeOfdata+=getc(OutputFileStream)<<24;
                printf("Data Size . %12d .\n",SizeOfdata);

                unsigned long CRC32=crc32_initialise();
                unsigned long AccuripCRC=crc32_initialise();
                unsigned char count;
                unsigned char Maxcount=0;
                unsigned long Samples=0;
                unsigned long FirstKeyFrame=0;
                unsigned char BLANK=0;
                unsigned long TotalSamples=SizeOfdata/4;

                for (int Tmp=DriveOffset;Tmp<0;Tmp++)
                {
                    AccuripCRC=crc32_addchar(BLANK,AccuripCRC);
                    AccuripCRC=crc32_addchar(BLANK,AccuripCRC);
                    AccuripCRC=crc32_addchar(BLANK,AccuripCRC);
                    AccuripCRC=crc32_addchar(BLANK,AccuripCRC);                
                };


                while (SizeOfdata>0)
                {
                    
                    // First Sample
                    
                    long int DataReadSize;

                    if (SizeOfdata>BufferSize)
                        DataReadSize=BufferSize;
                    else
                        DataReadSize=SizeOfdata;

                    SizeOfdata-=DataReadSize;
                    
                    fread(Buffer,1,DataReadSize,OutputFileStream);

                    for (long int Offset=0;Offset<DataReadSize;Offset+=4)
                    {

                        CRC32=crc32_addchar(Buffer[0+Offset],CRC32);
                        CRC32=crc32_addchar(Buffer[1+Offset],CRC32);
                        CRC32=crc32_addchar(Buffer[2+Offset],CRC32);
                        CRC32=crc32_addchar(Buffer[3+Offset],CRC32);
                       
                        if (Samples>=DriveOffset && Samples<(TotalSamples+DriveOffset))
                        {
                            AccuripCRC=crc32_addchar(Buffer[0+Offset],AccuripCRC);
                            AccuripCRC=crc32_addchar(Buffer[1+Offset],AccuripCRC);
                            AccuripCRC=crc32_addchar(Buffer[2+Offset],AccuripCRC);
                            AccuripCRC=crc32_addchar(Buffer[3+Offset],AccuripCRC);
                        };
                        
                        if(Maxcount<15)
                        {
                            count=0;
                            if ((Buffer[0+Offset]&0xe0)==0x40) 
                            {
                                count=2;
                            };
                            

                            //MSB
                            if ((Buffer[1+Offset]&0xff)==0x15) 
                            {
                                count+=8;
                            };
                            

                            // Second Sample
                            if ((Buffer[2+Offset]&0xe0)==0x40)                    
                            {
                                count++;
                            };
                            
                            //MSB
                            if ((Buffer[3+Offset]&0xff)==0x15) 
                            {
                                count+=4;
                            };

                            if(count>Maxcount) 
                            {
                                Maxcount=count;
                                FirstKeyFrame=Samples;
                            }                                            
                        }
                        
                        Samples++;
                    }
                }

                for (int Tmp=DriveOffset;Tmp>0;Tmp--)
                {
                    AccuripCRC=crc32_addchar(BLANK,AccuripCRC);
                    AccuripCRC=crc32_addchar(BLANK,AccuripCRC);
                    AccuripCRC=crc32_addchar(BLANK,AccuripCRC);
                    AccuripCRC=crc32_addchar(BLANK,AccuripCRC);                
                };
                printf("Samples : %10d\n",Samples);            
                SampleOffset=FirstKeyFrame % 588;
                printf("First Key Frame : %d offset %d Mask %2d.\n",
                    (FirstKeyFrame / 588),
                        SampleOffset,
                            Maxcount);

                printf("CRC : %08x\n",crc32_read(CRC32));            
                printf("Accurip-CRC : %08x\n",crc32_read(AccuripCRC));            

            };
        };
        fclose(OutputFileStream);
    }
    else
    {
        printf("Could not open file");
    };
};


BOOL send_test_unit_ready(HANDLE hVolume)
{

    unsigned char cmd[]={0,0,0,0,0,0};
    BOOL success= SendCommandToDrive(hVolume ,
                            6,
                                SCSI_IOCTL_DATA_UNSPECIFIED,
                                    0,
                                        100,
                                            &DataBuf,
                                                cmd);

    unsigned char key, ASC, ASCQ;

    key = sptd_sb.SenseBuf[2] & 0xf;
    ASC = sptd_sb.SenseBuf[12];
    ASCQ = sptd_sb.SenseBuf[13];

    if(key==SEN_KEY_NOT_READY)
    {
        if(ASC==0x3A)
        {
            if(ASCQ==0x00)
            {
                //printf(" - Medium not present.");
            }
            else
            if(ASCQ==0x01)
            {
                //printf(" - Medium not present-tray closed.");
            }
            else
            if(ASCQ==0x02)
            {
                //printf(" - Medium not present-tray open.");
            }
        }
    };
    return success;
}


BOOL send_read_TOC_PMA_ATIP(HANDLE hVolume, 
                            DWORD in_data_trans_len, 
                                UCHAR in_format)
{


    unsigned char cmd[]={0x43,
                            0,
                                in_format,
                                    0,
                                        0,
                                            0,
                                                0,
                                                     (in_data_trans_len>> 8) & 0xff,
                                                        (in_data_trans_len) & 0xff,    
                                                            0};
    ZeroMemory(DataBuf, in_data_trans_len);

    return SendCommandToDrive(hVolume ,
                            10,
                                SCSI_IOCTL_DATA_IN,
                                    in_data_trans_len,
                                        100,
                                            &DataBuf,
                                                cmd);
};



/* Sends Read TOC/PMA/ATIP command to read TOC, check for errors & display it. */
bool read_TOC(HANDLE hVolume)
{

    DWORD i;
    BOOL success;
    DWORD alloc_len=0;
    ULONG sector;
    //UCHAR DataBuf2[64*1024]; //A temporary duplicate buffer.

    //Sends MMC READ TOC/PMA/ATIP command to drive to get header.
    success=send_read_TOC_PMA_ATIP(hVolume, 4, 0);

    if(success)
    {
        if(sptd_sb.sptd.ScsiStatus==STATUS_GOOD)
        {
            alloc_len=DataBuf[0] << 8;
            alloc_len=alloc_len | DataBuf[1];
        }
        else
        {
            {
                //printf("Command sent but returned with an unhandled code \n");
            }
        }
    }
    else
    {
        //printf("DeviceIOControl returned with failed status.\n");
    }
    
    if(alloc_len>0)
    {
        //Sends MMC READ TOC/PMA/ATIP command to drive to get full data.
        success=send_read_TOC_PMA_ATIP(hVolume, alloc_len+4, 0);
        if(success)
        {
            if(sptd_sb.sptd.ScsiStatus==STATUS_GOOD)
            {
                //Display TOC data (excludes MMC reserved parts)..
                //printf("\n");

                int TrackCount;
                TrackCount=0;

                UCHAR FirstSession;
                FirstSession=DataBuf[5];
                int start;

                for(i=4;i<alloc_len;i=i+8)
                {
                    //TOC ADR & Control byte.
                    //printf("%02d", DataBuf[i+1]);

                    if ((DataBuf[i+1]==FirstSession) && !(DataBuf[i+2]==0xAA))
                    {
                        //printf("%02d", DataBuf[i]);
                        //printf("%02d", DataBuf[i+1]);

                        //printf("Length ");
                        //Track number(hex).
                        //printf("%02d", DataBuf[i+2]);
                        
                        sector=DataBuf[i+4] << 24;
                        sector=sector | (DataBuf[i+5] << 16);
                        sector=sector | (DataBuf[i+6] << 8);
                        sector=sector | DataBuf[i+7];
                        TOC[TrackCount].StartSectorLBA =sector;

                        
                        //LBA sector address of TOC entry.
                        //printf("%08X", sector);
                        //printf(" ");
                       // printf("%d",sector);

                        start=sector;

                        sector=DataBuf[i+12] << 24;
                        sector=sector | (DataBuf[i+13] << 16);
                        sector=sector | (DataBuf[i+14] << 8);
                        sector=sector | DataBuf[i+15];
                        
                        if (DataBuf[i+9]!=FirstSession)
                                {sector=sector-11400;};
 
                        TOC[TrackCount].EndSectorLBA =sector;

                        //printf("%f",((sector-start)/75.0));
                        //printf("%c",10);

                        TOC[TrackCount].TrackNumber =DataBuf[i+2];
                        TrackCount++;
                    };
                }
                
            }
            else
            {
                {                  
                    //printf("Command sent but returned with an unhandled status code");
                }
            }
        }
        else
        {
           // printf("\n");
           // printf("DeviceIOControl returned with failed status.\n");
        };


    }
    else
    {
        success=false;
    }
    return success;
}




/* Sends Read TOC/PMA/ATIP command to read TOC, check for errors & display it. */
bool read_CDText(HANDLE hVolume)
{


//    DWORD i;
    BOOL success;
    DWORD alloc_len=0;
//    ULONG sector;
    //UCHAR DataBuf2[64*1024]; //A temporary duplicate buffer.

    printf("CDTEXT STARTS\n");

    //Sends MMC READ TOC/PMA/ATIP command to drive to get header.
    success=send_read_TOC_PMA_ATIP(hVolume, 4, CDROM_READ_TOC_EX_FORMAT_CDTEXT);

    if(success)
    {
        if(sptd_sb.sptd.ScsiStatus==STATUS_GOOD)
        {
            alloc_len=DataBuf[0] << 8;
            alloc_len=alloc_len | DataBuf[1];
        }
        else
        {
            {
                printf("CD Text Not Found \n");
            }
        }
    }
    else
    {
        printf("CD Text Not Found \n");
    }
    
    if(alloc_len>0)
    {

        //Sends MMC READ TOC/PMA/ATIP command to drive to get full data.

        success=send_read_TOC_PMA_ATIP(hVolume, alloc_len+4, CDROM_READ_TOC_EX_FORMAT_CDTEXT);

        if(success)
        {
        //move the data into the data buffer
            CDROM_TOC_CD_TEXT_DATA TheTextData;
            ZeroMemory(&TheTextData,sizeof(TheTextData));
            memcpy(&TheTextData,&DataBuf,alloc_len);
            if(sptd_sb.sptd.ScsiStatus==STATUS_GOOD)
            {
                //Display TOC data (excludes MMC reserved parts)..
                //printf("\n");

                int FrameCount=alloc_len / sizeof(TheTextData.Descriptors[0]);
                int start;
                UCHAR PacketType=0;
                int TrackNo;
                bool StartedOutput=false;
                for (int Ind=0 ; Ind<FrameCount ; Ind++ )
                {
                    if (TheTextData.Descriptors[Ind].PackType ==CDROM_CD_TEXT_PACK_GENRE)
                    {
                        int Genre=((TheTextData.Descriptors[Ind].Text[1] & 0x0f) +
                                        ((TheTextData.Descriptors[Ind].Text[1] >>4 )*10));

                        printf("Genre%c%2d%c%s\n",
                                    9,
                                        TheTextData.Descriptors[Ind].TrackNumber ,
                                            9,
                                                Genres[Genre]);
                    }
                    else
                    {
                        int CharactersinBuffer;
                        bool Unicode=(TheTextData.Descriptors[Ind].Unicode!=0);
                        if (TheTextData.Descriptors[Ind].Unicode==0)
                        {
                            CharactersinBuffer=12;
                        }
                        else
                        {
                            CharactersinBuffer=6;                    
                        };
                        if (PacketType!=TheTextData.Descriptors[Ind].PackType )
                        {
                            TrackNo=TheTextData.Descriptors[Ind].TrackNumber ;
                            PacketType=TheTextData.Descriptors[Ind].PackType;
                            if (StartedOutput)
                            {
                                    printf("\n");
                                    StartedOutput=false;
                            }
                        }
                        for (int InnerInd=0 ; InnerInd<CharactersinBuffer  ; InnerInd++ )
                        {
                            if (StartedOutput)
                            {
                                if ((!Unicode && TheTextData.Descriptors[Ind].Text[InnerInd]==0) 
                                        || (Unicode && 
                                            TheTextData.Descriptors[Ind].WText [InnerInd]==0))
                                {
                                    printf("\n");
                                    StartedOutput=false;
                                    TrackNo+=1;
                                };            
                            
                            }
                            else
                            {
                                //UCHAR PacketType;
                                if ((!Unicode && TheTextData.Descriptors[Ind].Text[InnerInd]>0) 
                                        || (Unicode && 
                                            TheTextData.Descriptors[Ind].WText [InnerInd]>0))
                                {
                                    //We have a new one Starting

                                    StartedOutput=true;
                                    switch( PacketType ) 
                                    {
                                        case CDROM_CD_TEXT_PACK_ALBUM_NAME:
                                            printf("Title%c%2d%c",9,TrackNo,9);
                                                break;
                                        case CDROM_CD_TEXT_PACK_PERFORMER:
                                            printf("Artist%c%2d%c",9,TrackNo,9);
                                                break;
                                        case CDROM_CD_TEXT_PACK_SONGWRITER:
                                            printf("Writer%c%2d%c",9,TrackNo,9);
                                                break;
                                        case CDROM_CD_TEXT_PACK_COMPOSER:
                                            printf("Composer%c%2d%c",9,TrackNo,9);
                                                break;
                                        case CDROM_CD_TEXT_PACK_ARRANGER:
                                            printf("Arranger%c%2d%c",9,TrackNo,9);
                                                break;
                                        case CDROM_CD_TEXT_PACK_MESSAGES:
                                            printf("Comment%c%2d%c",9,TrackNo,9);
                                                break;
                                        case CDROM_CD_TEXT_PACK_DISC_ID:
                                            printf("Cat No%c%2d%c",9,TrackNo,9);
                                                break;
                                        case CDROM_CD_TEXT_PACK_UPC_EAN:
                                            if (TrackNo == 0)
                                                {
                                                printf("UPC/EAN%c%c",9,9);
                                                }
                                            else
                                                {
                                                printf("ISRC%c%2d%c",9,TrackNo,9);
                                                }
                                            break;

                                        default:
                                            PacketType=0;
                                            StartedOutput=false;
                                    }
                                }                            
                            };
                       
                            if (StartedOutput)
                            {
                                if (Unicode)
                                {
                                    printf("%C",(TheTextData.Descriptors[Ind].WText[InnerInd]));
                                }
                                else
                                {
                                    if ((TheTextData.Descriptors[Ind].Text[InnerInd]!='-') 
                                                || (PacketType!=CDROM_CD_TEXT_PACK_UPC_EAN))
                                    {
                                        printf("%c",TheTextData.Descriptors[Ind].Text[InnerInd]);
                                    }
                                }
                            }
                        };
                    };               
                };
                
                 /* 
                  FILE * Temp;
                    start=fopen_s(&Temp,("c:\\JukeBoxData\\RawCDTEXT.raw") , "wb");
                    start=fwrite(&DataBuf[4],1,alloc_len,Temp);
                    fclose(Temp);
                 */
            }
            else
            {
               // printf("CD Text Not Found \n");
            }
        }
        else
        {
           // printf("CD Text Not Found \n");
        };
    }
    else
    {
        //printf("CD Text Not Found \n");
        success=false;
    }
    printf("CDTEXT ENDS\n");

    return success;
}


void OutputTOC()
{
    printf("TOC START\n");
    byte TrackCount=0;
    while ((TrackCount<100) && (TOC[TrackCount].TrackNumber>0))
    {
        //Track number(hex).
        printf("%02d%c", TOC[TrackCount].TrackNumber,9);

        //LBA sector address of TOC entry.
        printf("%d%c", TOC[TrackCount].StartSectorLBA,9);

        printf("%d%c",(TOC[TrackCount].EndSectorLBA-TOC[TrackCount].StartSectorLBA),9);

        printf("%f\n",((TOC[TrackCount].EndSectorLBA-TOC[TrackCount].StartSectorLBA)/75.0));
        
        TrackCount++;
    }
    printf("TOC END\n");
}


/* Sends Test Unit Ready command 3 times, check for errors & display error info. */
BOOL test_unit_ready3(HANDLE hVolume)
{
    unsigned char i=3;
    BOOL success;
    /*
    Before sending the required command, here we clear any pending sense info from the drive
    which may interfere by sending Test Unit Ready command at least 3 times if neccessary.
    ----------------------------------------------------------------------------------------*/
    do
    {            
        //Sends SPC1 Test Unit Ready command to drive
        success=send_test_unit_ready(hVolume);
        if(success)
        {
            if(sptd_sb.sptd.ScsiStatus==STATUS_GOOD)
            {
                i=1;
            };
        }
        i--;
    }
    while(i>0);
    return success;
}

void FindAndOutputUPC(HANDLE hVolume) 
{
    printf("UPC STARTS\n");
    if (ReadISRCorUPC(hVolume, 
          0, 
            SectorsPerRead, 
                1000, 
                    false))
        {
            printf("%s\n",DecodeUPC());
        };
    printf("UPC ENDS\n");

};

void FindAndOutputISRCs(HANDLE hVolume) 
{
    byte TrackCount;
    TrackCount=0; 
    printf("ISRCS STARTS\n");
    while ((TrackCount<100)&&(TOC[TrackCount].TrackNumber !=0))
    {
        printf("ISRC%c%02d%c",9,TOC[TrackCount].TrackNumber,9);
        if (ReadISRCorUPC(hVolume, 
              TOC[TrackCount].StartSectorLBA, 
                    SectorsPerRead, 
                        1000,true))
        {
            unsigned char *ToPrint=(DecodeISRCCode());
            printf("%12s",ToPrint);
        };
        printf("\n");
        TrackCount++;
    }
    printf("ISRCS ENDS\n");
}

void usage()
{    
    printf("usage:\n\nGet information From Disc\nmedia_info drive [letter] <toc> <upc> <isrc> <cdtext>\n");  
    printf("\nRip Track to Wav\nmedia_info drive [letter] readtrack [Track No] [Filename]\n"); 
    printf("\nCheck the CRC for a wav\nmedia_info checkcrc [Filename] <Acurip Offset in samples>\n\nSwitches:- [compulsory]  <optional>\n");
    return ;
}

int ParseArguments(int argc,char * argv[],char * SearchFor)
{
    byte Count;
    for (Count=1;Count<argc;Count++)
    {
        if (!strcmp(argv[Count],SearchFor))  
        {
            return Count;
        };
    };
    return 0;
};


int main(int argc, char * argv[])  // Array of environment variable strings
{
    BOOL success=true;
    HANDLE hVolume;
    //Open the volume handle to the drive.


    if (argc>1)
    {

        if (ParseArguments(argc,argv,"drive")> 0)
        {
            char DrivePos=ParseArguments(argc,argv,"drive");
            if (argc>DrivePos+2)
            { 
                hVolume = OpenVolume(argv[DrivePos+1][0]);

                if(hVolume == INVALID_HANDLE_VALUE)
                {
                    return false;
                }
                else
                {
                    try
                    {
                        if (test_unit_ready3(hVolume))
                        {

                            if (ParseArguments(argc,argv,"info")> 0)
                                printf("Drive Info : %s\n",scsi_inquiry(hVolume));
                            else
                            {
                                if (read_TOC(hVolume))
                                {
                                    if (ParseArguments(argc,argv,"toc")> 0)
                                        OutputTOC();

                                    if (ParseArguments(argc,argv,"cdtext")> 0)
                                        read_CDText(hVolume);

                                    if (ParseArguments(argc,argv,"upc")> 0)
                                        FindAndOutputUPC(hVolume);

                                    if (ParseArguments(argc,argv,"isrc")> 0)
                                        FindAndOutputISRCs(hVolume);

                                    if (ParseArguments(argc,argv,"readtrack")> 0)
                                    {
                                        int readtrackpos=ParseArguments(argc,argv,"readtrack");
                                        if (argc>readtrackpos+2)
                                        { 
                                            ReadTrackToOutputStream(hVolume,
                                            atoi(argv[readtrackpos+1])-1,argv[readtrackpos+2]);   
                                        };
                                    }
                                }
                                else
                                {
                                    printf("Unable to retrieve TOC\n");
                                };
                            }
                        };
                    }

                    catch (char * str)
                    {
                        printf("%s",str);
                        success=false; 
                    };

                // Close the volume handle to the drive.
                if(!CloseVolume(hVolume))
                    success= false;
                };
            }
        }
        else
        {        
            if (ParseArguments(argc,argv,"checkcrc")> 0)
            {
                int readtrackpos=ParseArguments(argc,argv,"checkcrc");
                if (argc>readtrackpos+1)
                { 
                    if (argc>readtrackpos+2)
                    {
                        GenerateCRC32ForTrack(argv[readtrackpos+1],atoi(argv[readtrackpos+2]));
                    }
                    else
                    {
                        GenerateCRC32ForTrack(argv[readtrackpos+1],0);
                    };
                };
            };
        };                
    }        
    else
    {
        usage();
        success=false;
    };
    return success;
};